import pathlib
import sys

import pytest
from pydantic import ValidationError


ROOT_DIR = pathlib.Path(__file__).resolve().parents[1]
if str(ROOT_DIR) not in sys.path:
    sys.path.insert(0, str(ROOT_DIR))


from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_root import EntityModelRoot


def _mixin_edl():
    return {
        "edl_version": "1.0",
        "escopo": "core",
        "id": "auditoria",
        "description": "Mixin de auditoria",
        "mixin": True,
        "properties": {
            "atualizado_em": {
                "type": "datetime",
            }
        },
        "repository": {
            "properties": {
                "atualizado_em": {"column": "lastupdate"},
            }
        },
    }


def _entity_edl():
    return {
        "edl_version": "1.0",
        "escopo": "core",
        "id": "empresa",
        "description": "Entidade real",
        "properties": {
            "id": {"type": "uuid", "pk": True},
        },
        "required": ["id"],
        "repository": {
            "properties": {
                "id": {"column": "id"},
            }
        },
        "api": {
            "resource": "empresas",
            "verbs": ["GET"],
            "expose": True,
        },
    }


def test_entity_model_root_accepts_mixin_without_repository_map():
    model = EntityModelRoot(**_mixin_edl())

    assert model.mixin is True
    assert model.repository.map is None


def test_entity_model_root_rejects_mixin_with_repository_map():
    edl = _mixin_edl()
    edl["repository"]["map"] = "core.auditoria"

    with pytest.raises(ValidationError, match="Mixins não devem definir"):
        EntityModelRoot(**edl)


def test_entity_model_rejects_non_mixin_without_repository_map():
    with pytest.raises(ValidationError, match="repository\\.map"):
        EntityModel(**_entity_edl())



class MissingEntityConfigException(Exception):
    pass


class MissingDependencyConfigException(Exception):
    """
    Erro de alto nível para falha de carga de dependências de uma entidade.
    """

    def __init__(
        self,
        *,
        entity_resource: str,
        dependency_resource: str,
        dependency_scope: str | None = None,
        dependency_tenant: int | str | None = None,
        dependency_grupo_empresarial: str | None = None,
    ) -> None:
        self.entity_resource = entity_resource
        self.dependency_resource = dependency_resource
        self.dependency_scope = dependency_scope
        self.dependency_tenant = dependency_tenant
        self.dependency_grupo_empresarial = dependency_grupo_empresarial
        super().__init__(self._build_message())

    @property
    def dependency_key(self) -> str:
        # Formato legível para mensagens de API e logs.
        if self.dependency_scope:
            return f"{self.dependency_scope}/{self.dependency_resource}"
        return self.dependency_resource

    def _build_message(self) -> str:
        return (
            f"Missing dependency config '{self.dependency_key}' "
            f"required by '{self.entity_resource}'."
        )



from __future__ import annotations

from typing import Dict, List, Literal, Optional

from pydantic import AliasChoices, BaseModel, Field, model_validator

APIVerbs = Literal["GET", "POST", "PUT", "DELETE", "PATCH"]


class HandlerMapping(BaseModel):
    attr: str = Field(..., description="Nome do atributo no tipo composto de destino.")
    from_: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("from", "from_"),
        serialization_alias="from",
        description="Path de origem no payload (ex.: body.estabelecimento).",
    )
    as_: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("as", "as_"),
        serialization_alias="as",
        description="Tipo composto do elemento quando houver arrays (ex.: ns.t_item[]).",
    )
    mapping: Optional[List["HandlerMapping"]] = Field(
        default=None,
        description="Sub-mapeamento utilizado quando 'from' aponta para coleções.",
    )


class HandlerArgBinding(BaseModel):
    type_name: str = Field(
        ...,
        description="Nome qualificado do tipo composto esperado pela função no banco.",
    )
    mapping: List[HandlerMapping] = Field(
        ...,
        description="Mapeamento dos campos do payload para os atributos do tipo composto.",
    )


class HandlerCall(BaseModel):
    arg_binding: HandlerArgBinding = Field(
        ...,
        description="Configuração de binding do payload para o tipo composto.",
    )


class HandlerResult(BaseModel):
    expected: Literal["empty", "entity_row", "custom_json", "partial_row"] = Field(
        default="empty",
        description="Define o formato esperado do retorno da função.",
    )
    properties: Optional[List[str]] = Field(
        default=None,
        description="Lista de propriedades a serem retornadas quando o resultado esperado for 'partial_row'.",
    )

    @model_validator(mode="after")
    def validate_partial_row(self):
        if self.expected == "partial_row" and not self.properties:
            raise ValueError(
                "result.properties é obrigatório quando result.expected for 'partial_row'."
            )
        return self


class HandlerError(BaseModel):
    sqlstate: Optional[str] = Field(
        default=None,
        description="Código SQLSTATE tratado explicitamente.",
    )
    http_status: Optional[int] = Field(
        default=None,
        description="Código HTTP retornado quando o SQLSTATE for encontrado.",
    )
    message: Optional[str] = Field(
        default=None,
        description="Mensagem alternativa retornada quando o erro for capturado.",
    )


class HandlerConfig(BaseModel):
    impl: Optional[Literal["pg_function"]] = Field(
        default=None,
        description="Somente funções PostgreSQL (pg_function) são suportadas neste estágio.",
    )
    function_ref: Optional[str] = Field(
        default=None,
        description="Referência qualificada para a função no banco (ex.: schema.fn_nome).",
    )
    call: Optional[HandlerCall] = Field(
        default=None,
        description="Configuração de chamada (binding de argumentos).",
    )
    result: HandlerResult = Field(
        default_factory=HandlerResult,
        description="Especificação do que se espera da função.",
    )
    errors: Optional[List[HandlerError]] = Field(
        default=None,
        description="Tratamento específico para SQLSTATE conhecidos.",
    )


class APIModel(BaseModel):
    resource: str = Field(
        ...,
        description="Nome do recurso REST (rota base dos endpoints; exemplo: 'clientes').",
    )
    expose: Optional[bool] = Field(
        default=True,
        description="Indica se a API deve ser exposta (padrão: True).",
    )
    verbs: Optional[List[APIVerbs]] = Field(
        default=["GET", "POST", "PUT", "DELETE", "PATCH"],
        description="Lista de verbos HTTP suportados pela API (padrão: todos).",
    )
    default_sort: Optional[List[str]] = Field(
        None,
        description="Lista de campos usados na ordenação padrão (padrão: se nada for fornecido, será usada, ao menos, a PK).",
    )
    handlers: Optional[Dict[str, HandlerConfig]] = Field(
        default=None,
        description="Mapeamento opcional de verbos para handlers implementados via funções pg_function.",
    )


APIModel.model_rebuild()



import os
import sys
from pathlib import Path

import pytest
from pydantic import ValidationError

ROOT_DIR = Path(__file__).resolve().parents[1]
if str(ROOT_DIR) not in sys.path:
    sys.path.insert(0, str(ROOT_DIR))

os.environ.setdefault("ESCOPO_RESTLIB2", "test-scope")

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel


def _base_edl() -> dict:
    return {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para campo lut.",
        "id": "LutEntity",
        "version": "1.0",
        "required": ["id"],
        "properties": {
            "id": {"type": "uuid", "pk": True},
        },
        "repository": {
            "map": "test.lut_entities",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "lut-entities",
            "expose": True,
            "verbs": ["GET"],
        },
    }


def test_entity_model_accepts_lut_boolean():
    edl = _base_edl()
    edl["lut"] = True

    model = EntityModel(**edl)

    assert model.lut is True


def test_entity_model_rejects_invalid_lut_value():
    edl = _base_edl()
    edl["lut"] = "talvez"

    with pytest.raises(ValidationError):
        EntityModel(**edl)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional


@dataclass
class FunctionRelationBinding:
    """
    Representa o vínculo entre um campo do DTO e um campo declarado
    no FunctionType correspondente.
    """

    field_name: str
    function_type_class: Optional[str] = None


@dataclass
class FunctionBindingConfig:
    """
    Estrutura usada pelo PropertyCompiler para aplicar metadados de
    binding entre DTOs e FunctionTypes (insert/update).
    """

    insert_fields: Dict[str, str] = field(default_factory=dict)
    update_fields: Dict[str, str] = field(default_factory=dict)
    insert_relations: Dict[str, FunctionRelationBinding] = field(
        default_factory=dict
    )
    update_relations: Dict[str, FunctionRelationBinding] = field(
        default_factory=dict
    )

    def merge_insert_field(self, dto_field: str, function_field: str) -> None:
        if not dto_field or not function_field:
            return
        self.insert_fields[dto_field] = function_field

    def merge_update_field(self, dto_field: str, function_field: str) -> None:
        if not dto_field or not function_field:
            return
        self.update_fields[dto_field] = function_field

    def merge_insert_relation(
        self, dto_field: str, binding: FunctionRelationBinding
    ) -> None:
        if not dto_field or not binding:
            return
        self.insert_relations[dto_field] = binding

    def merge_update_relation(
        self, dto_field: str, binding: FunctionRelationBinding
    ) -> None:
        if not dto_field or not binding:
            return
        self.update_relations[dto_field] = binding


@dataclass
class FunctionCompilationOutput:
    """
    Resultado da compilação de um FunctionType (insert/update).
    """

    class_name: Optional[str] = None
    code: Optional[str] = None
    function_name: Optional[str] = None
    field_bindings: Dict[str, str] = field(default_factory=dict)
    relation_bindings: Dict[str, FunctionRelationBinding] = field(
        default_factory=dict
    )



from __future__ import annotations

from typing import Any

from nsj_rest_lib2.compiler.edl_model.primitives import PrimitiveTypes, STR_BASED_TYPES


def parse_primitive_type(property_type: Any) -> PrimitiveTypes | None:
    """
    Converte valor de tipo para `PrimitiveTypes` quando aplicavel.

    Args:
        property_type: Valor bruto do campo `type`.

    Returns:
        Enum `PrimitiveTypes` quando reconhecido; caso contrario, None.
    """
    if isinstance(property_type, PrimitiveTypes):
        return property_type
    if not isinstance(property_type, str):
        return None
    normalized = property_type.strip().lower()
    if not normalized:
        return None
    try:
        return PrimitiveTypes(normalized)
    except Exception:
        return None


def is_primitive_property_type(property_type: Any) -> bool:
    """
    Indica se o tipo representa um primitivo canonico do EDL.

    Args:
        property_type: Valor do campo `type`.

    Returns:
        True quando `property_type` pertence a `PrimitiveTypes`.
    """
    return parse_primitive_type(property_type) is not None


def is_relation_property_type(property_type: Any) -> bool:
    """
    Indica se o tipo representa relacionamento (interno ou externo).

    Args:
        property_type: Valor do campo `type`.

    Returns:
        True para tipos de relacao (ex.: `escopo/recurso`, `#/components/...`).
    """
    if is_primitive_property_type(property_type):
        return False
    if not isinstance(property_type, str):
        return False
    normalized = property_type.strip()
    if not normalized:
        return False
    return "/" in normalized or normalized.startswith("#/")


def is_textual_primitive_type(property_type: Any) -> bool:
    """
    Indica se o tipo primitivo possui semantica textual.

    Args:
        property_type: Valor do campo `type`.

    Returns:
        True para tipos definidos em `STR_BASED_TYPES`.
    """
    primitive_type = parse_primitive_type(property_type)
    if primitive_type is None:
        return False
    return primitive_type in STR_BASED_TYPES


def resolve_effective_repository_column(
    prop_name: str,
    prop_meta: Any,
    repo_prop_meta: Any,
) -> str | None:
    """
    Resolve coluna fisica efetiva para uma propriedade de EDL.

    Args:
        prop_name: Nome logico da propriedade.
        prop_meta: Metadados da propriedade no schema.
        repo_prop_meta: Metadados em `repository.properties.<prop>`.

    Returns:
        Nome da coluna local quando aplicavel; None para relacoes sem coluna
        explicita.
    """
    if isinstance(repo_prop_meta, dict):
        column = repo_prop_meta.get("column")
    else:
        column = getattr(repo_prop_meta, "column", None)

    if column is not None and str(column).strip():
        return str(column).strip()

    if isinstance(prop_meta, dict):
        property_type = prop_meta.get("type")
    else:
        property_type = getattr(prop_meta, "type", None)

    if is_relation_property_type(property_type):
        return None

    return prop_name



import os
import sys
from pathlib import Path

ROOT_DIR = Path(__file__).resolve().parents[1]
if str(ROOT_DIR) not in sys.path:
    sys.path.insert(0, str(ROOT_DIR))

os.environ.setdefault("ESCOPO_RESTLIB2", "test-scope")
os.environ.setdefault("REDIS_HOST", "localhost")
os.environ.setdefault("REDIS_PORT", "6379")

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.model import CompilerResult
from nsj_rest_lib2.service.entity_config_writer import EntityConfigWriter


def _base_edl() -> dict:
    return {
        "edl_version": "1.0",
        "escopo": "folha",
        "description": "Entidade de teste para payload de entity_config.",
        "id": "LutEntity",
        "version": "1.0",
        "required": ["id"],
        "properties": {
            "id": {"type": "uuid", "pk": True},
        },
        "repository": {
            "map": "lookuptables.lut_entities",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "lut-entities",
            "expose": True,
            "verbs": ["GET"],
        },
    }


def test_build_payload_includes_lut_true():
    edl = _base_edl()
    edl["lut"] = True
    model = EntityModel(**edl)
    result = CompilerResult()

    payload = EntityConfigWriter("folha").build_payload(model, result)

    assert payload["lut"] is True


def test_build_payload_includes_lut_false_by_default():
    model = EntityModel(**_base_edl())
    result = CompilerResult()

    payload = EntityConfigWriter("folha").build_payload(model, result)

    assert payload["lut"] is False



from __future__ import annotations

from typing import Any

from nsj_rest_lib2.compiler.edl_model.primitives import PrimitiveTypes, PropertyType


class MigrationCompilerUtil:
    """
    Funções utilitárias para montagem de trechos SQL usados nas migrações.
    """

    SQL_TYPE_MAP = {
        PrimitiveTypes.STRING: "varchar(250)",
        PrimitiveTypes.TEXT: "varchar",
        PrimitiveTypes.NUMBER: "numeric(20,8)",
        PrimitiveTypes.INTEGER: "integer",
        PrimitiveTypes.BOOLEAN: "boolean",
        PrimitiveTypes.UUID: "uuid",
        PrimitiveTypes.CURRENCY: "numeric(20,4)",
        PrimitiveTypes.QUANTITY: "numeric(20,4)",
        PrimitiveTypes.CPF: "varchar(20)",
        PrimitiveTypes.CNPJ: "varchar(20)",
        PrimitiveTypes.CPF_CNPJ: "varchar(20)",
        PrimitiveTypes.EMAIL: "varchar(100)",
        PrimitiveTypes.DATE: "date",
        PrimitiveTypes.DATETIME: "timestamp with time zone",
        PrimitiveTypes.DURATION: "interval",
    }

    @staticmethod
    def resolve_sql_type(datatype: PropertyType, max_length: int | None) -> str:
        if not isinstance(datatype, PrimitiveTypes):
            raise ValueError(f"Tipo de propriedade não suportado para SQL: {datatype}")

        base_type = MigrationCompilerUtil.SQL_TYPE_MAP.get(datatype)
        if not base_type:
            raise ValueError(f"Tipo de propriedade não mapeado para SQL: {datatype}")

        if datatype in (PrimitiveTypes.STRING, PrimitiveTypes.TEXT) and max_length:
            return f"varchar({max_length})"

        return base_type

    @staticmethod
    def is_numeric(datatype: PropertyType) -> bool:
        if not isinstance(datatype, PrimitiveTypes):
            return False
        return datatype in (
            PrimitiveTypes.NUMBER,
            PrimitiveTypes.INTEGER,
            PrimitiveTypes.CURRENCY,
            PrimitiveTypes.QUANTITY,
        )

    @staticmethod
    def quote_literal(value: Any) -> str:
        if value is None:
            return "null"
        if isinstance(value, bool):
            return "true" if value else "false"
        if isinstance(value, (int, float)):
            return str(value)
        escaped = str(value).replace("'", "''")
        return f"'{escaped}'"

    @staticmethod
    def comment_on_column(table_name: str, column_name: str, description: str) -> str:
        escaped = description.replace("'", "''")
        return (
            f"COMMENT ON COLUMN {table_name}.{column_name} IS '{escaped}'"
        )

    @staticmethod
    def drop_constraint_if_exists(table_name: str, constraint_name: str) -> str:
        return f"ALTER TABLE {table_name} DROP CONSTRAINT IF EXISTS {constraint_name}"

    @staticmethod
    def check_constraint_name(table_name: str, column_name: str, suffix: str) -> str:
        safe_table = table_name.replace(".", "_")
        return f"{safe_table}_{column_name}_{suffix}"

    @staticmethod
    def _safe_table_identifier(table_name: str) -> str:
        # Constraint names não aceitam '.' como separador de schema.
        return table_name.replace(".", "_")

    @staticmethod
    def primary_key_constraint_name(table_name: str) -> str:
        return f"{MigrationCompilerUtil._safe_table_identifier(table_name)}_pkey"

    @staticmethod
    def resolve_default_sql(
        datatype: PropertyType, default_value: Any
    ) -> tuple[str | None, bool]:
        if default_value is None:
            return None, False

        # EDL legado usa "uuid.uuid4" como expressão Python.
        # No SQL, convertemos para expressão PostgreSQL sem depender de extensões extras.
        if isinstance(datatype, PrimitiveTypes) and datatype == PrimitiveTypes.UUID:
            if isinstance(default_value, str):
                normalized = default_value.strip().lower()
                if normalized in {"uuid.uuid4", "uuid4", "uuid.uuid1", "uuid1"}:
                    return "(md5(random()::text || clock_timestamp()::text)::uuid)", True

        return MigrationCompilerUtil.quote_literal(default_value), False

    @staticmethod
    def add_enum_check_constraint(
        table_name: str, column_name: str, values: list[Any]
    ) -> list[str]:
        statements: list[str] = []
        constraint_name = MigrationCompilerUtil.check_constraint_name(
            table_name, column_name, "enum_chk"
        )
        statements.append(
            MigrationCompilerUtil.drop_constraint_if_exists(
                table_name, constraint_name
            )
        )
        if values:
            formatted_values = ", ".join(
                MigrationCompilerUtil.quote_literal(v) for v in values
            )
            statements.append(
                f"ALTER TABLE {table_name} ADD CONSTRAINT {constraint_name} "
                f"CHECK ({column_name} IN ({formatted_values}))"
            )
        return statements

    @staticmethod
    def add_max_check_constraint(
        table_name: str, column_name: str, maximum: int | float
    ) -> list[str]:
        statements: list[str] = []
        constraint_name = MigrationCompilerUtil.check_constraint_name(
            table_name, column_name, "max_chk"
        )
        statements.append(
            MigrationCompilerUtil.drop_constraint_if_exists(
                table_name, constraint_name
            )
        )
        statements.append(
            f"ALTER TABLE {table_name} ADD CONSTRAINT {constraint_name} "
            f"CHECK ({column_name} <= {maximum})"
        )
        return statements

    @staticmethod
    def add_primary_key(
        table_name: str, column_names: list[str], constraint_name: str | None = None
    ) -> str:
        constraint = (
            constraint_name
            or MigrationCompilerUtil.primary_key_constraint_name(table_name)
        )
        columns = ", ".join(column_names)
        return (
            f"ALTER TABLE {table_name} "
            f"ADD CONSTRAINT {constraint} PRIMARY KEY ({columns})"
        )

    @staticmethod
    def drop_primary_key(table_name: str, constraint_name: str | None = None) -> str:
        constraint = (
            constraint_name
            or MigrationCompilerUtil.primary_key_constraint_name(table_name)
        )
        return f"ALTER TABLE {table_name} DROP CONSTRAINT IF EXISTS {constraint}"

    @staticmethod
    def indent_sql(sql: str, spaces: int) -> list[str]:
        prefix = " " * spaces
        return [prefix + line if line else prefix for line in sql.split("\n")]



import os
import sys
from pathlib import Path

import pytest
from pydantic import ValidationError

ROOT_DIR = Path(__file__).resolve().parents[1]
if str(ROOT_DIR) not in sys.path:
    sys.path.insert(0, str(ROOT_DIR))

os.environ.setdefault("ESCOPO_RESTLIB2", "test-scope")

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel


def _base_edl() -> dict:
    return {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para validacoes estruturais.",
        "id": "ValidationEntity",
        "version": "1.0",
        "required": ["id"],
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "codigo": {"type": "string"},
            "nome": {"type": "string"},
        },
        "repository": {
            "map": "test.validation_entities",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
                "codigo": {"column": "codigo"},
                "nome": {"column": "nome"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "validation-entities",
            "expose": True,
            "verbs": ["GET"],
        },
    }


def test_pk_property_must_be_required():
    edl = _base_edl()
    edl["required"] = []

    with pytest.raises(ValidationError) as exc:
        EntityModel(**edl)

    errors = exc.value.errors()
    assert any(
        error.get("type") == "pk_not_required" and error.get("loc") == ("required",)
        for error in errors
    )


def test_repository_column_collision_raises_validation_error():
    edl = _base_edl()
    edl["repository"]["properties"]["nome"]["column"] = "codigo"

    with pytest.raises(ValidationError) as exc:
        EntityModel(**edl)

    errors = exc.value.errors()
    assert any(
        error.get("type") == "duplicate_repository_column"
        and error.get("loc") == ("repository", "properties")
        for error in errors
    )


def test_relation_without_local_column_does_not_collide():
    edl = _base_edl()
    edl["properties"]["categoria"] = {
        "type": "custom/categoria_profissional",
        "cardinality": "1_1",
    }
    edl["repository"]["properties"]["categoria"] = {
        "relation_column": "custom/categoria_profissional/id"
    }

    model = EntityModel(**edl)

    assert "categoria" in model.properties



def format_entity_name(resource: str) -> str:
    """
    Formata o identificador tecnico do recurso para nome amigavel.

    Args:
        resource: Nome do recurso (ex.: "categorias-profissionais").

    Returns:
        Nome formatado para exibicao em mensagens de erro/log.
    """
    # Normaliza o nome tecnico do recurso para texto amigavel em mensagens.
    value = (resource or "").strip().replace("-", " ").replace("_", " ")
    if not value:
        return "Desconhecida"

    lower = value.lower()
    if lower.endswith("oes") and len(lower) > 3:
        lower = lower[:-3] + "ao"
    elif lower.endswith("aes") and len(lower) > 3:
        lower = lower[:-3] + "ao"
    elif lower.endswith("res") and len(lower) > 3:
        lower = lower[:-2]
    elif lower.endswith("s") and len(lower) > 1:
        lower = lower[:-1]

    return " ".join(word.capitalize() for word in lower.split())



from pydantic import Field, ValidationError, model_validator
from pydantic_core import PydanticCustomError

from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.validation_util import (
    resolve_effective_repository_column,
)


class EntityModelRoot(EntityModelBase):
    escopo: str = Field(..., description="Escopo do EDL (define a aplicação).")

    @model_validator(mode="after")
    def validate_repository_and_required_consistency(self) -> "EntityModelRoot":
        """
        Aplica validações estruturais do EDL na entidade raiz.

        Regras:
        - Propriedade PK precisa constar em `required`;
        - Coluna física não pode ser mapeada por mais de uma propriedade;
        - `repository.map` é obrigatório para entidades não mixin;
        - Mixins não devem definir `repository.map`.

        Returns:
            Próprio modelo validado.

        Raises:
            ValidationError: Quando houver inconsistências estruturais.
        """
        line_errors: list[dict] = []

        required_set = set(self.required or [])
        for prop_name, prop_meta in (self.properties or {}).items():
            if getattr(prop_meta, "pk", False) and prop_name not in required_set:
                line_errors.append(
                    {
                        "type": PydanticCustomError(
                            "pk_not_required",
                            "Propriedade PK '{property}' deve constar em required.",
                            {"property": prop_name},
                        ),
                        "loc": ("required",),
                        "input": self.required,
                    }
                )

        repo_properties = self.repository.properties or {}
        column_to_props: dict[str, list[str]] = {}
        for prop_name, prop_meta in (self.properties or {}).items():
            repo_prop_meta = repo_properties.get(prop_name)
            effective_column = resolve_effective_repository_column(
                prop_name=prop_name,
                prop_meta=prop_meta,
                repo_prop_meta=repo_prop_meta,
            )
            if not effective_column:
                continue
            column_to_props.setdefault(effective_column, []).append(prop_name)

        for column_name, prop_names in column_to_props.items():
            if len(prop_names) <= 1:
                continue
            props_sorted = ", ".join(sorted(prop_names))
            line_errors.append(
                {
                    "type": PydanticCustomError(
                        "duplicate_repository_column",
                        (
                            "Coluna física '{column}' mapeada por mais de uma propriedade: "
                            "{properties}."
                        ),
                        {
                            "column": column_name,
                            "properties": props_sorted,
                        },
                    ),
                    "loc": ("repository", "properties"),
                    "input": self.repository.properties,
                }
            )

        repository_map = (
            self.repository.map.strip()
            if isinstance(self.repository.map, str)
            else None
        )
        if self.mixin:
            if repository_map:
                line_errors.append(
                    {
                        "type": PydanticCustomError(
                            "mixin_repository_map_not_allowed",
                            "Mixins não devem definir repository.map.",
                            {},
                        ),
                        "loc": ("repository", "map"),
                        "input": self.repository.map,
                    }
                )
        elif not repository_map:
            line_errors.append(
                {
                    "type": PydanticCustomError(
                        "repository_map_required",
                        "repository.map é obrigatório para entidades não mixin.",
                        {},
                    ),
                    "loc": ("repository", "map"),
                    "input": self.repository.map,
                }
            )

        if line_errors:
            raise ValidationError.from_exception_data(
                self.__class__.__name__,
                line_errors,
            )

        return self



import uuid

from nsj_rest_lib.decorator.dto import DTO
from nsj_rest_lib.descriptor.dto_field import DTOField
from nsj_rest_lib.descriptor.dto_field_validators import DTOFieldValidators
from nsj_rest_lib.dto.dto_base import DTOBase


def _validate_service_account(dto_field: DTOField, value):
    if value is None:
        return None

    if not isinstance(value, str):
        raise ValueError(
            f"{dto_field.storage_name} deve ser do tipo string. Valor recebido: {value}."
        )

    value = value.strip()

    if not value.startswith("serv_acc_"):
        raise ValueError(
            f"{dto_field.storage_name} deve comecar com \"serv_acc_\". Valor recebido: {value}."
        )

    return DTOFieldValidators().validate_email(dto_field, value)


@DTO()
class EscopoDTO(DTOBase):

    id: uuid.UUID = DTOField(
        pk=True,
        resume=True,
        not_null=True,
        default_value=uuid.uuid4,
        strip=True,
        min=36,
        max=36,
        validator=DTOFieldValidators().validate_uuid,
    )

    codigo: str = DTOField(
        resume=True,
        not_null=True,
        strip=True,
        min=1,
        max=100,
        unique="escopo_codigo",
        candidate_key=True,
    )

    service_account: str = DTOField(
        resume=True,
        strip=True,
        min=5,
        max=320,
        validator=_validate_service_account,
    )



import json

from typing import Any, Callable

from flask import Flask, request

from nsj_rest_lib.settings import APP_NAME, get_logger

from nsj_gcf_utils.rest_error_util import format_json_error

from nsj_multi_database_lib.decorator.multi_database import multi_database

from nsj_rest_lib.controller.controller_util import DEFAULT_RESP_HEADERS
from nsj_rest_lib.controller.delete_route import DeleteRoute
from nsj_rest_lib.controller.get_route import GetRoute
from nsj_rest_lib.controller.list_route import ListRoute
from nsj_rest_lib.controller.patch_route import PatchRoute
from nsj_rest_lib.controller.post_route import PostRoute
from nsj_rest_lib.controller.put_route import PutRoute

from nsj_rest_lib2.exception import (
    MissingDependencyConfigException,
    MissingEntityConfigException,
)
from nsj_rest_lib2.service.entity_loader import EntityLoader
from nsj_rest_lib2.util.entity_name_util import format_entity_name

logger = get_logger()


def _get_query_args() -> tuple[str | None, str | None, bool]:
    """
    Le tenant/grupo_empresarial/force_reload da query string e do body.

    Returns:
        Tupla com:
            - tenant (str | None)
            - grupo_empresarial (str | None)
            - force_reload (bool)
    """
    # Tentando ler do query args
    query_args = request.args
    tenant = query_args.get("tenant")
    grupo_empresarial = query_args.get("grupo_empresarial")
    force_reload = query_args.get("force_reload", "false").lower() == "true"

    # Tenta ler do corpo da requisicao quando nao vier por query string.
    try:
        body_str = request.data.decode("utf-8")
        body_json = json.loads(body_str)

        if not tenant:
            tenant = body_json.get("tenant")
        if not grupo_empresarial:
            grupo_empresarial = body_json.get("grupo_empresarial")
    except:
        pass

    # Preserva None quando nao houver dado para evitar "None" como string.
    tenant_value = str(tenant) if tenant is not None else None
    grupo_value = str(grupo_empresarial) if grupo_empresarial is not None else None
    return (tenant_value, grupo_value, force_reload)


def _endpoint_name(func: Any, multidb: bool, root: str) -> str:
    suffix = "_mb" if multidb else ""
    return f"{root}_{func.__name__}{suffix}"


def _route_handle(route: Any, *args: Any, **kwargs: Any) -> Any:
    if hasattr(route, "internal_handle_request"):
        return route.internal_handle_request(*args, **kwargs)
    return route.handle_request(*args, **kwargs)


def setup_dynamic_routes(
    flask_app: Flask,
    multidb: bool = True,
    dynamic_root_path: str = "edl1",
    injector_factory: Any = None,
    escopo_in_url: bool = False,
    edls_path: str | None = None,
    on_missing_entity_config: (
        Callable[[str, str, str | None, str | None], bool] | None
    ) = None,
    before_dynamic_request: (
        Callable[[dict[str, Any]], tuple[Any, int, dict[str, Any]] | None] | None
    ) = None,
) -> None:
    """
    Registra rotas dinamicas (GET/POST/PUT/PATCH/DELETE) para EDLs carregados.

    Args:
        flask_app: Instancia Flask onde as rotas serao registradas.
        multidb: Define se aplicara decorator de multi database por service account.
        dynamic_root_path: Prefixo de rota dinamica.
        injector_factory: Fabrica de injecao usada pelas rotas.
        escopo_in_url: Quando True, exige escopo na URL dinamica.
        edls_path: Caminho local de EDLs (fallback/disco).
        on_missing_entity_config: Callback para tentativa de recuperar/compilar
            config quando nao encontrada em cache/redis.
        before_dynamic_request: Callback executado antes da carga dinamica.
            Se retornar resposta HTTP, interrompe o fluxo da rota.

    Returns:
        None.
    """

    if not escopo_in_url:
        COLLECTION_DYNAMIC_ROUTE = f"/{APP_NAME}/{dynamic_root_path}/<entity_resource>"
        ONE_DYNAMIC_ROUTE = f"/{APP_NAME}/{dynamic_root_path}/<entity_resource>/<id>"
    else:
        COLLECTION_DYNAMIC_ROUTE = (
            f"/{APP_NAME}/{dynamic_root_path}/<entity_escopo>/<entity_resource>"
        )
        ONE_DYNAMIC_ROUTE = (
            f"/{APP_NAME}/{dynamic_root_path}/<entity_escopo>/<entity_resource>/<id>"
        )

    entity_loader = EntityLoader(edls_path=edls_path)

    def _dynamic_route_wrapper(
        route_builder: Callable[[tuple[Any, ...]], Callable[..., Any]],
        endpoint_suffix: str | None = None,
    ) -> Callable[..., Any]:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if "entity_resource" not in kwargs:
                msg = "Faltando parametro identificador da entidade na URL."
                return (format_json_error(msg), 400, {**DEFAULT_RESP_HEADERS})
            entity_resource = kwargs.pop("entity_resource")

            entity_escopo = kwargs.pop("entity_escopo", "")

            tenant, grupo_empresarial, force_reload = _get_query_args()
            entity_id = kwargs.get("id")

            if before_dynamic_request:
                # Permite bloquear o request dinamico antes do load/execucao da rota.
                guard_response = before_dynamic_request(
                    {
                        "method": request.method,
                        "path": request.path,
                        "host_url": request.host_url,
                        "entity_escopo": entity_escopo,
                        "entity_resource": entity_resource,
                        "entity_id": entity_id,
                        "tenant": tenant,
                        "grupo_empresarial": grupo_empresarial,
                        "force_reload": force_reload,
                    }
                )
                if guard_response is not None:
                    return guard_response

            try:
                entity_config = entity_loader.load_entity_source(
                    entity_resource,
                    tenant,
                    grupo_empresarial,
                    escopo=entity_escopo,
                    force_reload=force_reload,
                )
            except MissingEntityConfigException:
                recovered = False
                if on_missing_entity_config:
                    try:
                        recovered = bool(
                            on_missing_entity_config(
                                entity_resource,
                                entity_escopo,
                                tenant,
                                grupo_empresarial,
                            )
                        )
                    except Exception as exc:
                        logger.warning(
                            "Falha ao executar fallback de compilacao da entidade "
                            f"'{entity_escopo}/{entity_resource}': {exc}",
                            exc_info=True,
                        )
                if recovered:
                    try:
                        entity_config = entity_loader.load_entity_source(
                            entity_resource,
                            tenant,
                            grupo_empresarial,
                            escopo=entity_escopo,
                            force_reload=True,
                        )
                    except MissingEntityConfigException:
                        msg = (
                            f"Configuracao da entidade '{entity_resource}' nao encontrada."
                        )
                        return (
                            format_json_error(msg),
                            412,
                            {**DEFAULT_RESP_HEADERS},
                        )
                else:
                    msg = (
                        f"Configuracao da entidade '{entity_resource}' nao encontrada."
                    )
                    return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})
            except MissingDependencyConfigException as exc:
                # Mensagem funcional para diferenciar falta da entidade raiz
                # de falta de uma dependencia transitiva.
                entity_name = format_entity_name(exc.entity_resource)
                dependency_name = exc.dependency_key
                msg = (
                    f'Ocorreu um erro ao processar a entidade "{entity_name}" '
                    f'devido a ausencia de uma dependencia: "{dependency_name}"'
                )
                return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})

            (
                _,
                _,
                _,
                _,
                _,
                service_account,
                *rest,
            ) = entity_config

            route_handler = route_builder(entity_config)

            if multidb:
                route_handler = multi_database(service_account)(route_handler)

            return route_handler(*args, **kwargs)

        wrapper.__name__ = endpoint_suffix or f"{route_builder.__name__}_wrapper"
        return wrapper

    def list_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            _update_function_class_name,
            _insert_function_name,
            _update_function_name,
            _get_function_name,
            list_function_name,
            _delete_function_name,
            _get_function_type_class_name,
            list_function_type_class_name,
            _delete_function_type_class_name,
            *_,
            _custom_json_get_response,
            custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        def list_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "GET" not in api_verbs:
                return ("", 405, {})

            route = ListRoute(
                url=COLLECTION_DYNAMIC_ROUTE,
                http_method="GET",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                list_function_name=list_function_name,
                list_function_type_class=etities_dict.get(
                    list_function_type_class_name
                ),
                custom_json_response=bool(custom_json_list_response),
            )

            return _route_handle(route, *args, **kwargs)

        return list_dynamic

    def get_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            _update_function_class_name,
            _insert_function_name,
            _update_function_name,
            get_function_name,
            _list_function_name,
            _delete_function_name,
            get_function_type_class_name,
            _list_function_type_class_name,
            _delete_function_type_class_name,
            *_,
            custom_json_get_response,
            _custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        def get_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "GET" not in api_verbs:
                return ("", 405, {})

            route = GetRoute(
                url=ONE_DYNAMIC_ROUTE,
                http_method="GET",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                get_function_name=get_function_name,
                get_function_type_class=etities_dict.get(get_function_type_class_name),
                custom_json_response=bool(custom_json_get_response),
            )

            return _route_handle(route, *args, **kwargs)

        return get_dynamic

    def post_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            insert_function_class_name,
            _update_function_class_name,
            insert_function_name,
            _update_function_name,
            _get_function_name,
            _list_function_name,
            _delete_function_name,
            _get_function_type_class_name,
            _list_function_type_class_name,
            _delete_function_type_class_name,
            retrieve_after_insert,
            _retrieve_after_update,
            _retrieve_after_partial_update,
            post_response_dto_class_name,
            _put_response_dto_class_name,
            _patch_response_dto_class_name,
            custom_json_post_response,
            _custom_json_put_response,
            _custom_json_patch_response,
            _custom_json_get_response,
            _custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        def post_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "POST" not in api_verbs:
                return ("", 405, {})

            insert_function_type_class = (
                etities_dict.get(insert_function_class_name)
                if insert_function_class_name
                else None
            )

            dto_response_class = (
                etities_dict.get(post_response_dto_class_name)
                if post_response_dto_class_name
                else None
            )

            route = PostRoute(
                url=COLLECTION_DYNAMIC_ROUTE,
                http_method="POST",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                insert_function_type_class=insert_function_type_class,
                insert_function_name=insert_function_name,
                dto_response_class=dto_response_class,
                retrieve_after_insert=retrieve_after_insert,
                custom_json_response=bool(custom_json_post_response),
            )

            return _route_handle(route, *args, **kwargs)

        return post_dynamic

    def put_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            update_function_class_name,
            _insert_function_name,
            update_function_name,
            _get_function_name,
            _list_function_name,
            _delete_function_name,
            _get_function_type_class_name,
            _list_function_type_class_name,
            _delete_function_type_class_name,
            _retrieve_after_insert,
            retrieve_after_update,
            _retrieve_after_partial_update,
            _post_response_dto_class_name,
            put_response_dto_class_name,
            _patch_response_dto_class_name,
            _custom_json_post_response,
            custom_json_put_response,
            _custom_json_patch_response,
            _custom_json_get_response,
            _custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        dto_response_class = (
            etities_dict.get(put_response_dto_class_name)
            if put_response_dto_class_name
            else None
        )

        def put_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "PUT" not in api_verbs:
                return ("", 405, {})

            update_function_type_class = (
                etities_dict.get(update_function_class_name)
                if update_function_class_name
                else None
            )

            route = PutRoute(
                url=ONE_DYNAMIC_ROUTE,
                http_method="PUT",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                update_function_type_class=update_function_type_class,
                update_function_name=update_function_name,
                dto_response_class=dto_response_class,
                retrieve_after_update=retrieve_after_update,
                custom_json_response=bool(custom_json_put_response),
            )

            return _route_handle(route, *args, **kwargs)

        return put_dynamic

    def patch_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            _update_function_class_name,
            _insert_function_name,
            _update_function_name,
            _get_function_name,
            _list_function_name,
            _delete_function_name,
            _get_function_type_class_name,
            _list_function_type_class_name,
            _delete_function_type_class_name,
            _retrieve_after_insert,
            _retrieve_after_update,
            retrieve_after_partial_update,
            _post_response_dto_class_name,
            _put_response_dto_class_name,
            patch_response_dto_class_name,
            _custom_json_post_response,
            _custom_json_put_response,
            custom_json_patch_response,
            _custom_json_get_response,
            _custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        def patch_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "PATCH" not in api_verbs:
                return ("", 405, {})

            dto_response_class = (
                etities_dict.get(patch_response_dto_class_name)
                if patch_response_dto_class_name
                else None
            )

            route = PatchRoute(
                url=ONE_DYNAMIC_ROUTE,
                http_method="PATCH",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                dto_response_class=dto_response_class,
                retrieve_after_partial_update=retrieve_after_partial_update,
                custom_json_response=bool(custom_json_patch_response),
            )

            return _route_handle(route, *args, **kwargs)

        return patch_dynamic

    def delete_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            _update_function_class_name,
            _insert_function_name,
            _update_function_name,
            _get_function_name,
            _list_function_name,
            delete_function_name,
            _get_function_type_class_name,
            _list_function_type_class_name,
            delete_function_type_class_name,
            *_,
            _custom_json_get_response,
            _custom_json_list_response,
            custom_json_delete_response,
        ) = entity_config

        def delete_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "DELETE" not in api_verbs:
                return ("", 405, {})

            route = DeleteRoute(
                url=ONE_DYNAMIC_ROUTE,
                http_method="DELETE",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                delete_function_name=delete_function_name,
                delete_function_type_class=etities_dict.get(
                    delete_function_type_class_name
                ),
                custom_json_response=bool(custom_json_delete_response),
            )

            return _route_handle(route, *args, **kwargs)

        return delete_dynamic

    list_dynamic = _dynamic_route_wrapper(
        list_dynamic_builder, endpoint_suffix="list_dynamic"
    )
    get_dynamic = _dynamic_route_wrapper(
        get_dynamic_builder, endpoint_suffix="get_dynamic"
    )
    post_dynamic = _dynamic_route_wrapper(
        post_dynamic_builder, endpoint_suffix="post_dynamic"
    )
    put_dynamic = _dynamic_route_wrapper(
        put_dynamic_builder, endpoint_suffix="put_dynamic"
    )
    patch_dynamic = _dynamic_route_wrapper(
        patch_dynamic_builder, endpoint_suffix="patch_dynamic"
    )
    delete_dynamic = _dynamic_route_wrapper(
        delete_dynamic_builder, endpoint_suffix="delete_dynamic"
    )

    # Registrando as rotas no flask
    flask_app.add_url_rule(
        COLLECTION_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(list_dynamic, multidb, dynamic_root_path),
        view_func=list_dynamic,
        methods=["GET"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(get_dynamic, multidb, dynamic_root_path),
        view_func=get_dynamic,
        methods=["GET"],
    )
    flask_app.add_url_rule(
        COLLECTION_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(post_dynamic, multidb, dynamic_root_path),
        view_func=post_dynamic,
        methods=["POST"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(put_dynamic, multidb, dynamic_root_path),
        view_func=put_dynamic,
        methods=["PUT"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(patch_dynamic, multidb, dynamic_root_path),
        view_func=patch_dynamic,
        methods=["PATCH"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(delete_dynamic, multidb, dynamic_root_path),
        view_func=delete_dynamic,
        methods=["DELETE"],
    )
